#!/usr/bin/perl 
#
# Wombat - Rsync Backup Tool, Version 1.0
#
# Copyright 2005 Poul E.J. Petersen
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

use POSIX qw(:signal_h :errno_h :sys_wait_h);
use threads;
use threads::shared;
use Thread::Conveyor;
use File::Rsync;
use Getopt::Long;
#use strict;
use Time::Local;
use Fcntl qw(:DEFAULT :flock);
use DB_File;

if (! -d "/var/lib/wombat" ) {
  mkdir "/var/lib/wombat";
}

# In order to make certain that we do not try to simultaneously
# backup the same target from multiple threads or processes
# we keep a target-lock hash in a shared file. The format is
#
# $target_lock{host}=1;

my $target_lock_f='/var/lib/wombat/target_lock';	
my $target_lock_l="${target_lock_f}.lck";
my %target_lock;

# Images are written to a directory that is named by the time the
# image was created. There are many reasons why this image time
# may not match the scheduled time (slow backup for new target, 
# failed target being retried, etc). However, we need the pruner
# to operate on the scheduled time so we don't accidentaly prune
# directories that were supposed to be saved. For this reason,
# we store a DB file that maps images to the time that the image
# was scheduled to be created. The format is:
#
# schedmap{$target:schedtime}=image_time
#
# Where all times are in seconds since epoch.

my $schedmap_f='/var/lib/wombat/schedmap';
my $schedmap_l="${schedmap_f}.lck";
my %schedmap;

my %timeranges=(
	"hour"=>"0-23",
	"wday"=>"0-6",
	"mday"=>"1-31",
	"yday"=>"1-366",
	"month"=>"1-12",
	"epoch"=>"1-",
);

# Get the local program name and the path it was
# invoked from
my $myname=(reverse(split "[\\/]",$0))[0];          #Name of this program...
$myname=~s/([^\.]+).*/$1/;
(my $root=$0)=~s|^(.*)/[^/]+$|$1|g;

# Get the local hostname
chomp(my $localhost=`hostname | cut -d\. -f1`);

# Define the thread-safe work queue where we will store sync jobs
# to run. The contents of a work "package" will be:
#
# $target		A string of the form host:/path/to/file
#			  or host::modules
# $set                  A pointer to a hash of conf settings

my $workQ = Thread::Conveyor->new({maxboxes=>undef});
my $pruneQ = Thread::Conveyor->new({maxboxes=>undef});

# Here we have the thread globals

my @errors=();          #Stores thread errors
my $workers_done=0;	# A count of the workers that have finished
my $pruners_done=0;	# A count of the pruners that have finished

share(@errors);
share($workers_done);
share($pruners_done);

# Read the conf file into @buffer, removing comments as we go

my @buffer=();

open CONF, "<${root}/${myname}.conf";
while (my $line=<CONF>) {
  $line=(split("#",$line))[0];
  $line=~s/\s+$//;
  push @buffer,$line;
}
close CONF;

# Clear out the initial settings and alias hashes.
# And load defaults

my %aliases=();
my %set=(
	workers=>"1",
	pruners=>"1",
	maxconnect=>"2",
	runhours=>"1-23",
	rundays=>"0-6",
	rcmd=>"ssh",
);
	
# We need to record the scheduled start-time of this job, not the actual
# start time, in case there were delays getting the script running
# due to system resources, etc. This is especially important when
# computing boundary conditions against image times for retry and
# pruning. Starttime is also used to record when a job was 
# scheduled to run (as opposed to when it actually ran)

my @ltime=localtime(time());
$ltime[0]=0;	# set seconds to 0
$ltime[1]=0;	# set minutes to 0

$set{starttime}=timelocal(@ltime);	#The time this process started
$set{schedtime}=$set{starttime};	#The default schedule time for all jobs

# Now we parse the buffer and push jobs onto the queue

while ($#buffer > -1) {

  prep_buf(\@buffer);
 
  expand_alias(\%aliases,\@buffer);
  
  parse_alias(\%aliases,\@buffer) && next;
  parse_sync(\%set,\%aliases,\@buffer) && next;
  parse_set(\%set,\@buffer) && next;

  shift @buffer;
}

# Now we start up the worker and pruner threads

for (my $i=1;$i<=$set{pruners};$i++) {
  my $thr=threads->new(\&pruner);
  $thr->detach();
}

for (my $i=1;$i<=$set{workers};$i++) {
  my $thr=threads->new(\&worker);
  $thr->detach();
}

# Now we wait for the workers to finish. 

while ( $workers_done ne $set{workers} ) {
  sleep 3 
}
 
print "Workers are DONE\n";
$workQ->clean_dontwait;
$workQ->shutdown();

# Now we have to wait for the pruners to finish. The
# pruners use the Thread::Conveyor->take() method, so we
# signal that they should stop by pushing a NULL job onto
# the pruneQ

for (my $i=1;$i<=$set{pruners};$i++) {
  $pruneQ->put(undef);
}

while ( $pruners_done ne $set{pruners} ) {
  sleep 3 
}

$pruneQ->clean_dontwait;
$pruneQ->shutdown();

# END OF MAIN - Exit

#
# This subroutine processes a buffer and extracts the next largest block
# We expect to be passed two array pointers:
#
# \@buffer, \@block
#
# The results will be returned in \@block
#

sub block_extract {

  my $buffer=$_[0];
  my @block=();
  my $line="";

  my $nested="";        #We want the "largest" block, so we have to watch for nested blocks

  $line=read_buf($buffer);	#read_buf removes empty line
 
  if ($line!~/^\s*{/) { 	#there has to be a { or there is no block to extract
    unshift @{$buffer},$line;
    return;
  }

  while ($line ne "" && $nested ne "0") {
    while ($line=~/[{}]/ && $nested ne "0") {
      if ($line=~/^([^{}]*){(.*)/) {
        push @block,$1;
        $nested++;
        if ($nested > 1) {
          push @block,"{";
        }
        $line="$2";
      }
      if ($line=~/^([^{}]*)}(.*)/) {
        $nested--;
        push @block,$1;
        if ($nested > 0) {
          push @block,"}";
        }
        $line="$2";
      }
    }

    if ($nested ne "") {
      if ($nested > 0) { 
        push @block,"$line"; 
      } else { 
        unshift @{$buffer},$line;
      }
    } 

    if ($#{@{$buffer}} == -1) { last }
    $line=read_buf($buffer);

  }

  if ($nested ne 0) {
    { lock(@errors);
      push @errors,"Bad conf: missing a brace in a block statement...";
    }
  }

  unshift @{$buffer},$line;     #$line may still have contents, so push them back

# Return our copy of the decoded block in $_[1]

  @{$_[1]}=@block;

}

#
# This subroutine processes a buffer and extracts a list as bounded
# by (). Multi-dimensional lists are *NOT* supported.
# We expect to be passed two array pointers:
#
# \@buffer, \@list
#
# The results will be returned in \@list
#
# Elements in a list can be separated with newlines or commas.
# Commas can be included in a list element by quoting the element.
# Parens can be included in a list element by escaping the paren
#

sub list_extract {

  my $buffer=$_[0];
  my $list="";
  my $line="";
  my $remainder="";

  my $done=0;
  $line=read_buf($buffer);

  while ($list eq "" ) {
    $line=~s/\\\(/__escaped_leftparen__/g;
    $line=~s/\\\(/__escaped_rightparen__/g;

    if ($line=~/\([^\(\)]*\(/) {
      lock(@errors);
      push @errors,"Bad conf: multidimensional list are not supported...";
      return;
    }
    if ($line=~/^\s*\(([^\)]+)\)(.*)/) {
      $list=$1;
      $remainder=$2;

      $list=~s/__escaped_leftparen__/\\\(/g;
      $list=~s/__escaped_rightparen__/\\\)/g;
      $remainder=~s/__escaped_leftparen__/\\\(/g;
      $remainder=~s/__escaped_rightparen__/\\\)/g;

      unshift @{$buffer}, $remainder;
    } else {
      $line.=",".read_buf($buffer);
    }
  }

# Clean up white space and NULL fields
  my @list=();
  while ($list ne "") {
    push @list,word_extract(\$list);
  }

#  ($list=join(",",split /[\s,]+/,$list))=~s/^,(.*)/$1/;
#  @list=split ",",$list;

# Return our copy of the decoded list in $_[1]

  @{$_[1]}=@list;

}

# This subroutine extracts the next word from a string.
# Words are separated by whitespace unless they are quoted.

sub word_extract {

  my $string=$_[0];
  my $word="";

  $$string=~s/^[\s,]+(.*)/$1/;

  if (substr($$string,0,1) eq '"') {
    if ($$string=~/\"([^\"]+)\"(.*)/) {
      $word=$1;
      $$string=$2;
    } else {
      print "Error parsing next word from $$string\n";
    }
  }
  elsif ( $$string=~/^([^\s,]+)[\s,]*(.*)/) {
      $word=$1;
      $$string=$2;
  }

  return $word;
}
 
# Expand any aliases in the next line
sub expand_alias {

  (my $aliases, my $buffer)=@_;

  my $alias;
  foreach $alias (keys %{$aliases}) {
    ${@{$buffer}}[0]=~s/$alias/${%{$aliases}}{$alias}/g;
  }

  expand_newlines($buffer);
}

# Some commands replace a single line in the buffer with multiple
# lines. This sub fixes up the buffer by separating each line into
# an element.

sub expand_newlines {
  (my $buffer)=@_;
  my $line=shift @{$buffer};

  my @line=split /\n/,$line;
  @{$buffer}=( @line, @{$buffer} );

}

# This subroutine expands the next line from the buffer in place and does any preprocessing
# neccessary, for example processing exec commands or ignoring blank lines.

sub prep_buf {

  my $buffer=$_[0];

  if ($#{@{$buffer}} eq -1) {return}

# shift out any lines of whitespace

  while ( $#{@{$buffer}} ne -1 && ${@{$buffer}}[0] =~ /^\s*$/ ) { shift @{$buffer} };
  expand_exec($buffer);

# The expand might have added lines of whitespace, so shift them out again

  while ( $#{@{$buffer}} ne -1 && ${@{$buffer}}[0] =~ /^\s*$/ ) { shift @{$buffer} };

}

#
# This subroutine calls pre_buf on the buffer and then returns the next
# line, removing it from the buffer

sub read_buf {

  my $buffer=$_[0];
  if ($#{@{$buffer}} eq -1) {return}

  prep_buf($buffer);
  return shift @{$buffer};
}

# Expand any exec's in the line
sub expand_exec {

  (my $buffer)=@_;
  my $line=shift @{$buffer};

  if ($line =~ /(.*)exec((\s+\{.*)|(\{.*))/ ) {
    my $preexec=$1;
    unshift @{$buffer},$2;

    my @block=();
    block_extract($buffer,\@block);
    my $cmd=join(" ",@block);

    my $result_line="";
    my @result=( "$preexec" );

    open CMD, "$cmd 2>/dev/null |";

    while ( $result_line=<CMD> ) {
      $result_line=(split("#",$result_line))[0];
      $result_line=~s/\s+$//;
      push @result, $result_line;
    }

    @{$buffer}=( @result, @{$buffer} );

  } else {
    unshift @{$buffer},$line;
  }
}

# Process an alias command
sub parse_alias {
  (my $aliases, my $buffer)=@_;
  my $line=read_buf($buffer);
  my $alias_name;
  my $alias_list;

  if ($line =~ /^\s*alias/) {

# First, the {block} form

    if ($line =~ /^\s*alias\s*(_[^\ ]+)\s*(({.*)|())$/) {
      $alias_name=$1;
      unshift @{$buffer},$2;

      my @block=();
      block_extract($buffer,\@block);

      ${%{$aliases}}{"$alias_name"}=join("\n",@block);
    }

# The (list) form
    elsif ($line =~ /^\s*alias\s*(_[^\ ]+)\s*((\(.*)|())$/) {
      my $alias_name=$1;
      my $alias_list="";
      unshift @{$buffer},$2;

      my @list=();
      list_extract($buffer,\@list);

      ${%{$aliases}}{"$alias_name"}="(".join(",",@list).")";
    }

# The "string" form
    elsif ($line =~ /^\s*alias\s*(_[^\ ]+)\s*(.*)$/) {
      $alias_name=$1;
      ${%{$aliases}}{"$1"}="$2";
    }
    else {
      { lock(@errors);
        push @errors, "Bad conf parsing alias: $line\n";
      }
    }

    return 1;

  } else {
    unshift @{$buffer},$line;
    return 0;
  }
}

sub parse_set{
 
  (my $set, my $buffer) = @_;

  my $line=read_buf($buffer);

  if ($line=~/^\s*([^\s=]+)[\s=]+([^\(\{]+)/) {
    $set->{$1}=$2;
    return 1;
  } elsif ($line=~/^\s*([^\s]+)\s*(\(.*)/) {
    my $var=$1;
    unshift @{$buffer}, $2;
    my @list=();
    list_extract($buffer,\@list);
    $set->{$var}=\@list;
    return 1;
  } elsif ($line=~/^\s*([^\s]+)\s*(\{.*)/) {
    my $var=$1;
    unshift @{$buffer}, $2;
    my @block=();
    block_extract($buffer,\@block);
    $set->{$var}=\@block;
    return 1;
  } else {
    unshift @{$buffer},$line;
    return 0;
  }
}

sub parse_sync{
 
  (my $set, my $aliases, my $buffer) = @_;

  my $line=read_buf($buffer);
  my @targets=();
  my @block=();

  my $target;

  if ($line=~/^\s*sync\s*(\(.*)/) {
    unshift @{$buffer}, $1;

    list_extract($buffer,\@targets);
    block_extract($buffer,\@block);
 
    my %set=%{$set};
    my %aliases=%{$aliases};

    while ( $#block > -1 ) {
      prep_buf(\@block);
      expand_alias(\%aliases,\@block);
      parse_set(\%set,\@block) && next;
      shift @block;
    }
 
    foreach $target (@targets) {
      $workQ->put($target,\%set);
    }

    return 1;
  } else {
    unshift @{$buffer},$line;
    return 0;
  }
}

sub worker {
 
job: while ( (my $target,my $set)=$workQ->take_dontwait() ) {   	#get a new job until Q empty

  my %set=%{$set};	# Pull the data structure into a local hash

  print "opened job for $target\n";
  (my $host, my $path)=split_target($target);

# First thing we need to do is see if the run* settings require this
# target to be backed up at this time. If the job is not scheduled
# to be run at this time, then we need to check if the last run
# was missed. If not, then we loop for the next job effectively disposing this one.

  if ( ! ( check_interval($set{runhours},'0-23',(localtime($set{starttime}))[2]) &&
       check_interval($set{rundays},'0-7',(localtime($set{starttime}))[6] )) ) {
 
    print "Job $target not scheduled to run, checking for a missed image\n";

    my $dec;
    my $hrs_to_last_sch=0;

# Here we look back over the retry interval and see if a previous job was scheduled 
# within that time frame.

    for ($dec=1;$dec<=$set->{retry};$dec++) {
      if ( check_interval($set{runhours},'0-23',(localtime($set{starttime}-3600*$dec))[2] ) &&
           check_interval($set{rundays},'0-7',(localtime($set{starttime}-3600*$dec))[6] )) {
        $hrs_to_last_sch=$dec;
        last;
      }
    }  

#If we didn't find a scheduled job in the retry interval, then next

    if (! $hrs_to_last_sch) { 
      print "No scheduled jobs in retry interval ($set->{retry} hours). \n";
      next job; 
    } 
    
# As long as a completed job exists before now and anytime after the last scheduled
# job time, then we don't need to retry.     
     
    my $last_jobtime=localtime($set{starttime}-3600*$hrs_to_last_sch);

# We'll need the list of existing images for this target across all
# writeto paths. Each entry in @images is of the form [writeto_index,image_name]
# sorted by image_name most recent first.

    my @images=image_list($set{writeto},$target);
    my $itime=image_time($images[0][1]);

    if ($itime->{epoch} >= $set{starttime}-3600*$hrs_to_last_sch)  {
      print "$images[0][1] satisfies last scheduled jobtime ($last_jobtime). DONE\n";
      next job;
    }
 
    print "Retrying $target due to missed image at $last_jobtime\n";
    $set{schedtime}=$set{starttime}-3600*$hrs_to_last_sch;
  }
  
# We want to run a backup either because it is scheduled or the last
# scheduled job failed, but there are two reasons we might not run now:
#
#  1) The target is currently locked 	  in this case we just give up as long as the lock
#					    was acquired within maxruntime
#  2) The number of threads backing up
#     targets on host has met maxconnect  in this case we put this job back on the queue and try again later

  open LOCK, ">$target_lock_l";
  flock(LOCK,LOCK_EX);
  tie %target_lock, DB_File, "$target_lock_f";

  if ( $target_lock{"$target"} ) {
    print "Trying to make an image, but $target is locked. Skipping.\n";
    
    untie %target_lock;
    flock(LOCK,LOCK_UN);
    close(LOCK);

    my @images=image_list($set{writeto},$target);

    open LOCK, ">$schedmap_l"; 
    flock(LOCK,LOCK_EX);
    tie %schedmap, DB_File, "$schedmap_f";
 
    print "Substituting $images[0][1] for this image\n";
    $schedmap{"$target:".localtime_str($set{schedtime})}=$images[0][1];

    untie %schedmap;
    flock(LOCK,LOCK_UN);
    close LOCK;

    next job;
  }

  my $tcount=0;
  foreach my $tlock (keys %target_lock) {
    (my $hlock, my $plock)=split_target($tlock);
    if ($hlock eq $host) { $tcount++ }
  }

  if ($tcount>=$set{maxconnect} ) {
    
    print "Maxconnect reached trying to image $target\n";
    untie %target_lock;
    flock(LOCK,LOCK_UN);
    close(LOCK);

    if ( ! $set{wait} ) {
      print "Putting $target back on the queue in favor of other work\n";
      $set{wait}=1;
      $workQ->put($target,\%set);
      next job;

    } else {

      print "Already picked $target from the queue twice - waiting instead\n";

      while ($set{wait}) {
        sleep 15;

  	open LOCK, ">$target_lock_l";
  	flock(LOCK,LOCK_EX);
        tie %target_lock, DB_File, "$target_lock_f";

        my $tcount=0;
        foreach my $tlock (keys %target_lock) {
          (my $hlock, my $plock)=split_target($tlock);
          if ($hlock eq $host) { $tcount++ }
        }

        if ($tcount<$set{maxconnect}) {
          $set{wait}=0;
          $target_lock{"$target"}=1;
        }

        untie %target_lock;
    	flock(LOCK,LOCK_UN);
    	close(LOCK);
      }
    }
  } else { $target_lock{"$target"}=1 }

  untie %target_lock;
  flock(LOCK,LOCK_UN);
  close(LOCK);

# At this point we are ready to run the backup. However, it is possible that while
# we were getting ready (or waiting on tcount) that another process has made an image
# more recent than the one we are trying to fulfill. If so, we discard this job
# and give up our tcount hold.
# We'll need to reload the list of images to make certain it is current.

  @images=image_list($set{writeto},$target);
  my $itime=image_time($images[0][1]);
  
  if ($itime->{epoch} >= $set{starttime}) {
    print "Skipping $target since a newer image ($images[0][1]) exists\n";

    open LOCK, ">$target_lock_l";
    flock(LOCK,LOCK_EX);
    tie %target_lock, DB_File, "$target_lock_f";
    
    delete $target_lock{"$target"};
   
    untie %target_lock;
    flock(LOCK,LOCK_UN);
    close(LOCK);

    open LOCK, ">$schedmap_l"; 
    flock(LOCK,LOCK_EX);
    tie %schedmap, DB_File, "$schedmap_f";
 
    print "Substituting $images[0][1] for this image\n";
    $schedmap{"$target:".localtime_str($set{schedtime})}=$images[0][1];

    untie %schedmap;
    flock(LOCK,LOCK_UN);
    close LOCK;
    
    next job;
  }
  
# Now we are ready to process the job.
# In order to figure out what writeto dir to use, we get the index of the *last*
# writeto ($images[0][0]) increase this by one, and mod by the length of
# writeto. 

  my $next_writeto_index;

  if ($images[0][0] eq "") {
    $next_writeto_index=0 ;
  } else {
    $next_writeto_index=($images[0][0]+1)%($#{@{$set{writeto}}}+1);
  }
 
# The new image will be written to a subdirectory that represents the current
# time, which we get from localtime. We need to fix the year field and the mday 
# field. Then replace any single digit fields "n" with "0n" to make string comparison easy.

  my @now=localtime(time());
  $now[5]+=1900;
  $now[4]+=1;
  @now=map { if (length($_) eq 1 ) { "0$_" } else {"$_"} } @now;
  $set{image_time}="$now[5]-$now[4]-$now[3]_$now[2]:$now[1]";

# Now that we know the directory that we'll be working in and the new image name
# we need to know the most recent image in next_writeto to link against (if one exists).

  my $index="";
  my $last_image="";

  for ($index=0;$index<=$#images;$index++) {
    if ($images[$index][0] eq $next_writeto_index) {
      $last_image=$images[$index][1];
      last;
    }
  }

# We're ready to make the new image! If the test above found a previous image,
# we link against that using --link-dest. Otherwise, we just make a new image.

  my $write_path="$set{writeto}->[$next_writeto_index]/$host/$path/$set{image_time}";
  
  if ( ! -d "$set{writeto}->[$next_writeto_index]/$host/$path" ) {
    system "mkdir -p $set{writeto}->[$next_writeto_index]/$host/$path";
  }

  my %opts=(
    src=>"$target",
    dest=>"$write_path",
    archive=>"1",
    "one-file-system"=>"1",
    verbose=>1,
    verbose=>1,
    verbose=>1,
  );
#    delete=>"1", 

# If the target is formated as host:/path and a rcmd has been
# specified, then we need to add the rcmd options to %opts
# We don't add these settings for modules like host::module
# since these rsync modes are orthogonal (using a rsync 
# server is contrary to using a rcmd)
# Additionaly, if the target is not specified as a directory, 
# we add a trailing slash...

  if ($opts{src}=~/^[^:]+:\//) {
    if ($set{rcmd} eq "ssh" && $set{rcmd_user} ne "" ) {
      $opts{rsh}="ssh -l $set{rcmd_user}";
    } else {
      $opts{rsh}=$set{rcmd};
    }

    if ($opts{src}!~/\/$/) {
      $opts{src}.="/";
    }
  }

  if ( $last_image ne "" ) {
    $opts{"link-dest"}="../$last_image" ;
  }

# Handle SMB backups with smbmount

  my $localdir="";

  if ($set{smb}) {
    (my $smbhost, my $smbshare, my $smbpath) = split /:\//,$target,3;
   
    $localdir="/tmp/${myname}.${host}.${path}";
    system "mkdir $localdir";

    $opts{src}="${localdir}/${smbpath}/";

    system "mount -t smb -o ro,username=$set{smb}->[0],password=$set{smb}->[1] //$smbhost/$smbshare $localdir";

  }

  my $rsync=File::Rsync->new(\%opts);
  print join(" ",@{$rsync->getcmd})."\n";
  my $result=$rsync->exec;

# If the job fails critically, we substitute the last good image
# for this scheduled image.

  if ($result) { 
    print "Rsync image of $target was good\n";
  
    open LOCK, ">$schedmap_l"; 
    flock(LOCK,LOCK_EX);
    tie %schedmap, DB_File, "$schedmap_f";
 
    print "Setting schedmap for $target\n";
    $schedmap{"$target:".localtime_str($set{schedtime})}=$set{image_time};

    untie %schedmap;
    flock(LOCK,LOCK_UN);
    close LOCK;

  } else {
    print "Rsync of $target had errors!\n---\n".join(" ",@{$rsync->err})."---\n";

    if (! -d "$write_path" ) {

      print "Failure of $target was critical\n";
      open LOCK, ">$schedmap_l"; 
      flock(LOCK,LOCK_EX);
      tie %schedmap, DB_File, "$schedmap_f";
 
      print "Substituting $images[0][1] for this image\n";
      $schedmap{"$target:".localtime_str($set{schedtime})}=$images[0][1];

      untie %schedmap;
      flock(LOCK,LOCK_UN);
      close LOCK;
    }
  }

  if ($set{smb}) {
    system "umount $localdir"; 
    system "rmdir $localdir";
  }

# The backup is done, so we remove the target lock and add this target to the pruneQ

  open LOCK, ">$target_lock_l";
  flock(LOCK,LOCK_EX);
  tie %target_lock, DB_File, "$target_lock_f";

  delete $target_lock{"$target"};

  untie %target_lock;
  flock(LOCK,LOCK_UN);
  close LOCK;

# If the rsync returned no errors of it it at least worked well enough
# to create an image, then we prune this host.

  if ($result || -d "$write_path") {
    $pruneQ->put($target,\%set);
  }

  } #loop for another job

# This worker is done
  { lock($workers_done);
    $workers_done++;
  }
}


# This subroutine takes as input three values:
#
# $interval	A string representing an interval of integers.
#		  The interval can be defined as a comma separate list,
#		  a hyphen separate ranged, an asterisk, a negation, and a step
#		  limit on any of the above. Examples:
#		  1-10,12	 1 thru 10 and 12
#		  */2		 All values in $range, step 2
#		  1-10/2	 1 thru 10, step 2
#		  !10		 Every integer *except* 10
#		  !2-12		 Any outside the 2-12 range
#		  !2-12/2	 Any outside the 2-12 range, in steps of 2
#				   so 1,3,5,7,9,11,13,14,15,16, etc
#		  2-		 Match anything >= 2
#		  -2		 Match anything <= 2
#
# $maxrange	The range of allowable values if * or ! is used. Should
#		  be specified as "min-max". This will also restrict
#		  other interval specifications.
#
# $test		The element to test against the interval.

sub check_interval {

  (my $interval, my $maxrange, my $test) = @_;
  my $int="";
  my $range="";
  my $step="";
  my $val="";
 
# We use $maxrange to define the meaning of * in an interval
# specification and we also use it to bound possible values
# so we need the endpoints

  (my $min, my $max)=split /-/,$maxrange;

  $interval=~s/\*/$maxrange/g;	#replace * with $maxrange

# Each "interval spec" is separated with ","

  foreach my $int (split /\,/,$interval) {

# First see if we have a step declaration, otherwise the step=1

    if ( $int=~/([\d-\!]+)\/(\d+)/ ) {
      $range=$1;
      $step=$2;
    } else { 
      $range=$int;
      $step=1;
    }
    
# Now we have eight possibilities: single integer, a range a-b
# a range a- or a range -b and the negation of these four.
# and the negation of those two. I know there is an easier way to do
# this (rather than test all eight cases) but this is easier to read :)

    if ($range=~/^(\d+)-$/) {
      if ($test >= $range) { return 1 }
    } 
    elsif ($range=~/\!-(\d+)/) {
      if ($test > $range) { return 1 }
    } 
    elsif ($range=~/^-(\d+)/) {
      if ($test <= $range) { return 1}
    }
    elsif ($range=~/\!(\d+)-$/) {
      if ($test < $range) { return 1}
    }
    elsif ($range=~/^\!(\d+)-(\d+)/) {
      for ($val=$1;$val<=$2;$val+=$step) {
        if ($val <= $max && $val >= $min && $val eq $test) { return 0 }
      }
      return 1;
    } 
    elsif ($range=~/(\d+)-(\d+)/) {
      for ($val=$1;$val<=$2;$val+=$step) {
        if ($val <= $max && $val >= $min && $val eq $test) { return 1 }
      }
    }
    elsif ( $range=~/^\!(\d+)/ && $range ne $test ) { return 1 }
    elsif ( $range eq $test && $range <= $max && $range >= $min ) { return 1 }
  }
  
# If we get to here, then there were no matches.

  return 0;
}
   
# This sub generates a list of backup images for the specified target, sorted
# by creation date newest first. We expect the following inputs:
#
#   \@writeto		A pointer to an array of writeto paths
#   $target		The $host:$path spec
#   
# We return a sorted list of records formatted as 
#
# REC= [ writeto_index,		The index in the writeto array where this image is located
#        timestamp   ]		The timestamp of the image: year-month-day_hour:minute

sub image_list { 
  (my $writeto, my $target)=@_;
 
  (my $host, my $path)=split_target($target);

  my @images=();
  for (my $index=0;$index<=$#{@{$writeto}};$index++) {
    opendir DIR, "$writeto->[$index]/$host/$path";
    foreach my $image (grep { !/^\./ } readdir DIR) {
      push @images, ["$index","$image"];
    }
  }

  @images=sort { $b->[1] cmp $a->[1]; } @images;
  return @images;
}

# This sub takes as input a target ($host:$path) specification and returns
# the hostname and the pathname. Note that pathname is *not* the same as $path
# since $path may contain directory elements (/blah/blah/blah) whereas pathname
# will be a single top-level dir created by replacing [\/] with underscores.

sub split_target {

  (my $target) = @_;
 
  (my $host, my $path) = split /:+/, $target;

  $path=~s/[\/\\]/_/g;

  return ($host,$path);

}

# This sub handles the pruning of old jobs based on
# the preserve settings

sub pruner {

job: while ( (my $target,my $set)=$pruneQ->take() ) { 

  if (!defined($set)) { exit }	#Exit if we get a null job

  my %set=%{$set};	# Pull the data structure into a local hash

  print "prune job for $target\n";
  (my $host, my $path)=split_target($target);

# Load the schedmap entries for this target into %map. This will
# map image_time to a list of scheduled times so we can preserve images that
# were substituted.

  open LOCK, ">$schedmap_l";
  flock(LOCK,LOCK_EX);
  tie %schedmap, DB_File, "$schedmap_f";

  my %map;

  foreach my $spec (keys %schedmap) {
    if ($spec=~/^$target:(.*)$/) {
      push @{$map{$schedmap{$spec}}},$1
    }
  }

  untie %schedmap;
  flock(LOCK,LOCK_UN);
  close LOCK;

# We need the current list of images across all writeto dirs
# %keep will be a hash of 

  my @images=image_list($set{writeto},$target);
  my %keep;

# Foreach preserve statement, we check all the images and mark
# the ones that match the preserve spec. If a given image fails
# to match, but the preserve spec specifies an "atleast" parameter
# then we keep the image if it matches everything *except* the 
# epoch (computed from the "for" statement).

  foreach my $preserve ( @{$set{preserve}} ) {
    next if ($preserve eq "");

    my $count=0;
    my $intervals=timetext("$preserve");

    foreach my $image (@images) {

      foreach my $sched (@{$map{$image->[1]}},$image->[1]) {

        my $image_time=image_time($sched);

        if ( check_all_intervals($intervals,$image_time) ) {
          if ($image->[1] eq $sched) {
            print "$image->[1] normally preserved by $preserve\n";
          } else {
            print "$image->[1] (substituted for $sched) normally preserved by $preserve\n";
          }
          $keep{"$image->[1]"}++;
          $count++;
        }

        elsif ($intervals->{atleast} ne "" && $count<$intervals->{atleast} ) {
          delete ${%{$image_time}}{epoch};
          if ( check_all_intervals($intervals,$image_time) ) {
            if ($image->[1] eq $sched) {
              print "$image->[1] save by atleast spec from $preserve\n";
            } else {
              print "$image->[1] (substituted for $sched) saved by atleast spec from $preserve\n";
            }
            $keep{"$image->[1]"}++;
            $count++;
          }
        } 
      }
    }
  }

  foreach my $key (keys %keep) {
    print "Keeping $key (saved by $keep{$key} rules)\n";
  }

  foreach my $image (@images) {
    if (!$keep{$image->[1]}) {
      print "Pruning $set{writeto}->[$image->[0]]/$host/$path/$image->[1]\n";
      system "rm -rf $set{writeto}->[$image->[0]]/$host/$path/$image->[1]\n";

      open LOCK, ">$schedmap_l";
      flock(LOCK,LOCK_EX);
      tie %schedmap, DB_File, "$schedmap_f";
   
      foreach my $sched (@{$map{$image->[1]}}) {
        print "Removing $target schedmap: $sched->$image->[1]\n";
        delete $schedmap{"$target:$sched"};
      }

      untie %schedmap;
      flock(LOCK,LOCK_UN);
      close LOCK;
    }
  }

  } # loop for another job 

# This pruner is done
  { lock($pruners_done);
    $pruners_done++;
  }
}  

# This sub takes a "freeform" english string describing
# a time interval and returns a hash:
#
# (hour=>hour_str,
#  wday=>wday_str,
#  yday=>yday_str,
#  mday=>mday_str,
#  mon=>month_str,
#  for=>for_in_secs,
#  atleast=>integer
# );

sub timetext {
 
  my $timetext=$_[0];

# The default is to preserve all images. Individual settings
# over-ride this behavior and reduce the number of images kept.

  my %result=%timeranges;
  my @specs=split /\s+/,$timetext;

  while ( my $el=shift @specs ) {

    if ($el eq "for") {
      my $nextel=shift @specs;
      if ($nextel=~/\d+$/) { $nextel.=shift @specs; }
      if ($nextel=~/\d+\w+/) {
        $result{epoch}=time()-str_to_sec($nextel)."-";
      }
      next;
    }
    
    $result{$el}=shift @specs;
  }

  return \%result;
}      
     
# Convert a string of the form "\d\w" to seconds
# Where \w is the first letter of hours, days, weeks,
# months or years.

sub str_to_sec {

  my $str=$_[0];
  my $result;

  my %keys=(
	"h"=>3600,
	"d"=>3600*24,
	"w"=>3600*24*7,
	"m"=>3600*24*31,
	"y"=>3600*24*365.25,
  );

  if ( $str=~/(\d+)\s*(\w)/ ) {
    $result=$1*$keys{$2};
  } else {
    $result=$1;
  }
  
  return $result;
}

# This sub takes as input two time structure hashes (keys: hour, wday, etc)
# as follows:
#
# \%intervals	Time interval specs
# \%totest	The fixed values to check against the intervals.
# 
# Only the elements specified in %totest are checked. If an element is not
# listed, it is assumed to match.

sub check_all_intervals	 {

  (my $intervals, my $totest) = @_;

  foreach my $key ( keys %{$totest} ) {
    if (! check_interval($intervals->{$key},$timeranges{$key},$totest->{$key}) ) { return 0 }
  }
  
  return 1;

}

# This sub converts an epoch into an image string, like:
#
# YYYY-MM-DD_HH:MM

sub localtime_str {

  (my $epoch) = @_;

  my @localtime=localtime($epoch);

  $localtime[5]+=1900;
  $localtime[4]++;

  for (my $i=0;$i<$#localtime;$i++) {
    if (length($localtime[$i]) eq 1) { $localtime[$i]="0$localtime[$i]" }
  }

  return "$localtime[5]-$localtime[4]-$localtime[3]_$localtime[2]:$localtime[1]";
}

# This sub converts an "image name" into a time structure hash. The image_name is 
# formatted as : 
#
# year-month-day_hour:min

sub image_time {

  (my $image) = @_;

  if ($image eq "") { return }
  (my $year, my $month, my $mday, my $hour, my $min)=split /[-_:]/,$image;

  $year-=1900;
  $month-=1;

  my $epoch=timelocal("0",$min,$hour,$mday,$month,$year);
  my @localtime=localtime($epoch);

  $localtime[5]+=1900;
  $localtime[4]++;

  my %result= (
	"hour"=>"$localtime[2]",
	"wday"=>"$localtime[6]",
	"mday"=>"$localtime[3]",
	"yday"=>"$localtime[7]",
	"month"=>"$localtime[4]",
	"epoch"=>"$epoch",
  );

  return \%result;
}
